'''Module containing Utility Functions for the Publication Documentation'''

import re
from multiprocessing.pool import ThreadPool

import requests

from .models import Author, Journal, Publication
from .sparql import queryPublication

from ..config import endpoint
from ..getters import get_items, get_properties
from ..queries import query_sparql_pool

def additional_queries(publication, choice, key, parameter, function):
    '''Additional MaRDI Portal and Wikidata SPARQL Queries instance like authors or journals'''
    # Setup SPARQL queries for MaRDI Portal / Wikidata
    mardi_query = queryPublication[key].format(*parameter['mardi'])
    wikidata_query = queryPublication[key].format(*parameter['wikidata'])

    # Get Information from MaRDI Portal / Wikidata
    results = query_sparql_pool(
        {
            'wikidata': (wikidata_query, endpoint['wikidata']['sparql']),
            'mardi': (mardi_query, endpoint['mardi']['sparql'])
        }
    )
    # Extract Information from MaRDI Portal / Wikidata
    mardi_info = function(results['mardi'])
    wikidata_info = function(results['wikidata'])

    # Add (missing) MaRDI Portal / Wikidata IDs to authors
    assign_id(
        getattr(
            publication[choice],
            key
        ),
        mardi_info,
        'mardi'
    )
    assign_id(
        getattr(
            publication[choice],
            key
        ),
        wikidata_info,
        'wikidata'
    )

def assign_id(entities, target, prefix):
    '''Function to assign an ID to an entity.'''
    for entity in entities:
        if not entity.id or entity.id in ('not found', 'no author found', 'no journal found'):
            for id_entity in target.values():
                if entity.label.lower() == id_entity.label.lower():
                    entity.id = f"{prefix}:{id_entity.id}"
                    entity.label = id_entity.label
                    entity.description = id_entity.description

def assign_orcid(publication, source, id_type = 'orcid'):
    '''Funcion to assign an ORCiD to an Author'''
    for author in publication[source].authors:
        if not author.orcid_id:
            for id_author in publication[id_type].values():
                if author.label == id_author.label:
                    author.orcid_id = id_author.orcid_id

def extract_authors(data):
    '''Function to extract Author Information from query results'''
    authors = {}
    if data:
        for idx, entry in enumerate(data[0].get('authorInfos', {}).get('value', '').split(" | ")):
            if entry:
                authors[idx] = Author.from_query(entry)
    return authors

def extract_journals(data):
    '''Function to extract Journal Information from query results'''
    journals = {}
    if data:
        for idx, entry in enumerate(data[0].get('journalInfos', {}).get('value', '').split(" | ")):
            if entry:
                journals[idx] = Journal.from_query(entry)
    return journals

def generate_label(data):
    '''Function to generate a Publication Label for MathAlgoDB'''
    name = year = title = ''
    if data.authors:
        name = data.authors[0].label.split(' ')[-1]
    if data.date:
        year = data.date[:4]
    if data.title:
        title = data.title
    return f'{name} ({year}) {title}'

def get_citation(doi):
    '''Function to get citation information from DOI'''
    publication = {}

    if not re.match(r'10.\d{4,9}/[-._;()/:a-z0-9A-Z]+', doi):
        return publication

    choice = None

    # Define MaRDI Portal / Wikidata / MathAlgoDB SPARQL Queries
    mardi_query = queryPublication['MaRDI']['DOI_FULL'].format(
        doi,
        **get_items(),
        **get_properties()
    )
    wikidata_query = queryPublication['Wikidata']['DOI_FULL'].format(
        doi
    )
    mathalgodb_query = queryPublication['PublicationMathAlgoDBDOI'].format(
        doi
    )

    # Get Citation Data from MaRDI Portal / Wikidata / MathAlgoDB
    results = query_sparql_pool(
        {
            'wikidata': (wikidata_query, endpoint['wikidata']['sparql-scholarly']),
            'mardi':(mardi_query, endpoint['mardi']['sparql']),
            'mathalgodb':(mathalgodb_query, endpoint['mathalgodb']['sparql'])
        }
    )

    # Structure Publication Information
    for key in ['mardi', 'wikidata', 'mathalgodb']:
        try:
            publication[key] = Publication.from_query(results.get(key))
        except:
            publication[key] = None

    if not (publication['mardi'] or publication['wikidata']):
        # If no Citation Data in KGs get information from CrossRef, DataCite, DOI, zbMath
        pool = ThreadPool(processes=4)
        results = pool.map(
            lambda fn: fn(doi),
            [
                get_crossref_data,
                get_datacite_data,
                get_zbmath_data,
                get_doi_data
            ]
        )
        for idx, source in enumerate(['crossref', 'datacite', 'zbmath', 'doi']):
            if results[idx].status_code == 200:
                source_func_name = f"from_{source}"
                source_func = getattr(Publication, source_func_name)
                publication[source] = source_func(results[idx])
            else:
                publication[source] = None
        # Get Authors assigned to publication from ORCID
        publication['orcid'] = {}
        response = get_orcids(doi)
        if response.status_code == 200:
            orcids = response.json().get('result')
            if orcids:
                for idx, entry in enumerate(orcids):
                    orcid_id = entry.get('orcid-identifier', {}).get('path', '')
                    response = get_author_by_orcid(orcid_id)
                    if response.status_code == 200:
                        orcid_author = response.json()
                        publication['orcid'][idx] = Author.from_orcid(orcid_author)
        # Add (missing) ORCID IDs to authors
        for choice in ['crossref', 'datacite', 'zbmath', 'doi']:
            if publication[choice]:
                assign_orcid(publication, choice)
                break
        else:
            choice = None
        # Additional for chosen information source
        if choice:
            # Check if Authors already in MaRDI Portal or Wikidata
            orcid_id = ' '.join(
                f'"{author.orcid_id}"' if author.orcid_id else '""'
                for author in publication[choice].authors
            )
            zbmath_id = ' '.join(
                f'"{author.zbmath_id}"' if author.zbmath_id else '""'
                for author in publication[choice].authors
            )
            properties = get_properties()
            if orcid_id and zbmath_id:
                additional_queries(
                    publication,
                    choice,
                    'authors', 
                    {
                        'mardi': [
                            orcid_id,
                            zbmath_id,
                            properties['ORCID iD'],
                            properties['zbMATH author ID'],
                            properties['Wikidata QID']
                        ],
                        'wikidata': [
                            orcid_id,
                            zbmath_id,
                            'P496',
                            'P1556',
                            ''
                        ],
                    },
                    extract_authors
                )
            # Check if Journal already in MaRDI Portal or Wikidata
            journal_id = publication[choice].journal[0].issn

            if journal_id:
                additional_queries(
                    publication,
                    choice,
                    'journal',
                    {
                        'mardi': [
                            journal_id,
                            properties['ISSN']
                        ],
                        'wikidata': [
                            journal_id,
                            'P236'
                        ],
                    },
                    extract_journals
                )

    return publication

def get_crossref_data(doi):
    '''Function to get Citation Information from Crossref'''
    return requests.get(
        f"{endpoint['crossref']['api']}{doi}",
        timeout = 5
    )

def get_datacite_data(doi):
    '''Function to get Citation Information from Datacite'''
    return requests.get(
        f"{endpoint['datacite']['api']}{doi}",
        timeout = 5
    )

def get_doi_data(doi):
    '''Function to get Citation Information from DOI'''
    return requests.get(
        f"{endpoint['doi']['api']}{doi}",
        headers = {"accept": "application/json"},
        timeout = 5
    )

def get_zbmath_data(doi):
    '''Function to get Citation Information from ZbMath'''
    return requests.get(
        f"{endpoint['zbmath']['api']}{doi}",
        timeout = 5
    )

def get_orcids(doi):
    '''Function to get ORCiD Information from ORCiD'''
    return requests.get(
        f"{endpoint['orcid']['api']}/search/?q=doi-self:{doi}",
        headers = {'Accept': 'application/json'},
        timeout = 5
    )

def get_author_by_orcid(orcid_id):
    '''Function to get Author Information by ORCiD'''
    return requests.get(
        f"{endpoint['orcid']['api']}/{orcid_id}/personal-details",
        headers = {'Accept': 'application/json'},
        timeout = 5
    )



'''Functions for OAuth2 Authorization'''

import logging
import copy
import re
import json
import time
import random

from urllib.parse import urlencode

import requests
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

from .helpers import replace_in_dict

logger = logging.getLogger(__name__)


class OauthProviderMixin:
    '''Class containing functions for the authorization, post, and
       callback related to the OAuth2 protocoll.'''
    def post(self, request, jsons=None):
        '''Start the Posting Process'''
        self.store_in_session(request, 'request', ('post', jsons))
        return self.authorize(request)

    def authorize(self, request):
        '''Authorize for Posting'''
        state = get_random_string(length=32)
        self.store_in_session(request, 'state', state)
        url = self.authorize_url + '?' + urlencode(self.get_authorize_params(request, state))
        return HttpResponseRedirect(url)

    def callback(self, request):
        '''Ensure the Callback'''
        if request.GET.get('state') != self.pop_from_session(request, 'state'):
            return self.render_error(
                request,
                _('OAuth authorization not successful'),
                _('State parameter did not match.')
            )

        # Exchange the authorization code for an access token
        url = self.token_url + '?' + urlencode(self.get_callback_params(request))
        response = requests.post(
            url,
            self.get_callback_data(request),
            auth=self.get_callback_auth(request),
            headers=self.get_callback_headers(request),
            timeout = 10
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error(
                'callback error: %s (%s)',
                response.content,
                response.status_code
            )
            raise e

        response_data = response.json()
        access_token = response_data.get('access_token')

        # Replay the original request with the new access token
        try:
            _method, jsons = self.pop_from_session(request, 'request')
            return self.perform_post(
                request,
                access_token,
                jsons
            )
        except ValueError:
            pass

        return self.render_error(
            request,
            _('OAuth authorization successful'),
            _('But no redirect could be found.')
        )

    def perform_post(self, request, access_token, jsons=None):
        """Perform the actual Post"""
        init = copy.deepcopy(jsons)
        keys = list(jsons.keys())

        while keys:
            for key in list(jsons.keys()):
                if key not in keys:
                    continue

                payload = jsons[key].get("payload")
                url = jsons[key].get("url")

                if re.search(r"Item\d{10}", json.dumps(payload)):
                    continue
                if re.search(r"Item\d{10}", json.dumps(url)):
                    continue

                try:
                    if not jsons[key]['id']:
                        jsons = self._post_data(key, jsons, access_token)
                    else:
                        jsons = replace_in_dict(jsons, key, jsons[key]['id'])
                except RuntimeError as err:
                    return self.render_error(
                        request,
                        _('Something went wrong'),
                        str(err)
                    )

                time.sleep(0.1)

                keys.remove(key)

        final = jsons
        return self.post_success(request, init, final)

    def _post_data(self, key, jsons, access_token):
        """Post data for a single key, handles both relations and items, 
           with retries and backoff."""
        item = jsons[key]

        # Skip existing relations
        if key.startswith('RELATION') and item.get('exists') == 'true':
            return jsons

        # Reuse a requests.Session() to improve performance
        session = getattr(self, "_session", None)
        if session is None:
            session = requests.Session()
            self._session = session

        url = item['url']
        payload = item['payload']
        headers = self.get_authorization_headers(access_token)
        response = None

        for attempt in range(1, 5 + 1):
            try:
                response = session.post(url, json=payload, headers=headers, timeout=30)
                response.raise_for_status()
                wait = 0.1 + random.uniform(0, 0.5)
                time.sleep(wait)
                return self._handle_response(response, key, jsons)

            except requests.exceptions.Timeout:
                wait = 1.5 ** attempt + random.uniform(0, 0.5)
                logger.warning(
                    f"Timeout while posting {key} "
                    f"(attempt {attempt}/5), retrying in {wait:.2f}s"
                )
                time.sleep(wait)
                continue

            except requests.exceptions.ConnectionError as exc:
                wait = 1.5 ** attempt + random.uniform(0, 0.5)
                logger.warning(
                    f"Connection error posting {key}: {exc} "
                    f"(attempt {attempt}/{5}), retrying in {wait:.2f}s"
                )
                time.sleep(wait)
                continue

            except requests.HTTPError as exc:
                resp = exc.response
                status = resp.status_code if resp is not None else "no_response"
                content = resp.text if resp else "n/a"

                if status == 429:  # Too Many Requests
                    retry_after = int(resp.headers.get("Retry-After", 5))
                    logger.warning(f"Rate limit hit for {key}, waiting {retry_after}s before retry")
                    time.sleep(retry_after)
                    continue

                if isinstance(status, int) and status >= 500:  # Server-side transient error
                    wait = 1.5 ** attempt + random.uniform(0, 0.5)
                    logger.warning(f"Server error {status} posting {key}, retrying in {wait:.2f}s")
                    time.sleep(wait)
                    continue

                if status == 422:
                    return self._handle_policy_violation(resp, key, jsons)

                logger.error(f"HTTP error posting {key}: {status} - {content}")
                raise RuntimeError(_("POST request failed")) from exc

            except requests.exceptions.RequestException as exc:
                logger.error(f"Unexpected request error while posting {key}: {exc}")
                raise RuntimeError(_("Unexpected network error")) from exc

        # If we reach this point, all retries failed
        logger.error(f"POST {url} failed after {5} retries")
        raise RuntimeError(_("POST request failed after multiple retries"))

    def _handle_response(self, response, key, jsons):
        """Handle POST response and update placeholders.
           Returns updated jsons or raises on error.
        """
        try:
            response.raise_for_status()
            jsons[key]['id'] = response.json()['id']
            return replace_in_dict(jsons, key, response.json()['id'])

        except requests.HTTPError as exc:
            if response.status_code == 422:
                return self._handle_policy_violation(response, key, jsons)

            logger.warning('post error: %s (%s)', response.content, response.status_code)
            raise RuntimeError(_("POST request failed")) from exc


    def _handle_policy_violation(self, response, key, jsons):
        """Handle data-policy-violation errors and return updated jsons"""
        error_json = response.json()
        if error_json.get("code") == "data-policy-violation":
            violation = error_json.get("context", {}).get("violation")
            if violation == 'item-label-description-duplicate':
                conflict_id = (
                    error_json["context"]
                    .get("violation_context", {})
                    .get("conflicting_item_id")
                )
                if conflict_id:
                    jsons[key]['id'] = conflict_id
                    return replace_in_dict(jsons, key, conflict_id)

        # No fix possible → just return unchanged jsons
        return jsons

    def render_error(self, request, title, message):
        '''Render Error'''
        return render(request, 'core/error.html', {'title': title, 'errors': [message]}, status=200)

    def get_session_key(self, key):
        '''Get Session Key'''
        return f'{self.class_name}.{key}'

    def store_in_session(self, request, key, data):
        '''Store in Session'''
        session_key = self.get_session_key(key)
        request.session[session_key] = data

    def get_from_session(self, request, key):
        '''Get from Session'''
        session_key = self.get_session_key(key)
        return request.session.get(session_key)

    def pop_from_session(self, request, key):
        '''Pop from Session'''
        session_key = self.get_session_key(key)
        return request.session.pop(session_key, None)

    def get_authorization_headers(self, access_token):
        '''Get Authorization Header'''
        return {'Authorization': f'Bearer {access_token}'}

    def get_authorize_params(self, request, state):
        '''Get Authorization Parameter'''
        raise NotImplementedError

    def get_callback_auth(self, request):
        '''Get Callback Authorization'''
        return None

    def get_callback_headers(self, request):
        '''Get Callback Headers'''
        return {'Accept': 'application/json'}

    def get_callback_params(self, request):
        '''Get Callback Parameters'''
        return {}

    def get_callback_data(self, request):
        '''Get Callback Data'''
        return {}

    def post_success(self, request, init, final):
        '''Post Success'''
        raise NotImplementedError



'''Module containing general Models for MaRDMO'''

from dataclasses import dataclass
from typing import Optional

@dataclass
class Relatant:
    '''Data Class For Relatant Items'''
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]

    @classmethod
    def from_query(cls, raw: str) -> 'Relatant':
        '''Generate Item From Query'''
        identifier, label, description = raw.split(" | ")
        return cls(
            id = identifier,
            label = label,
            description = description,
        )

    @classmethod
    def from_triple(cls, identifier: str, label: str, description: str) -> 'Relatant':
        '''Generate Item From Triple'''
        return cls(
            id = identifier,
            label = label,
            description = description,
        )

    @classmethod
    def from_msc(cls, identifier: str, label: str, description: str) -> 'Relatant':
        '''Generate Item from MSC'''
        return cls(
            id = f"msc:{identifier}",
            label = label,
            description = description,
        )

@dataclass
class RelatantWithClass:
    '''Data Class For Relatant Items With Class'''
    id: Optional[str]
    label: Optional[str]
    description: Optional[str]
    item_class: Optional[str]

    @classmethod
    def from_query(cls, raw: str) -> 'RelatantWithClass':
        '''Generate Item From Query'''
        raw_split = raw.split(" | ")
        if len(raw_split) == 3:
            item_class = None
        else:
            item_class = raw_split[3]
        return cls(
            id = raw_split[0],
            label = raw_split[1],
            description = raw_split[2],
            item_class = item_class
        )



def compare_items(old, new):
    ids = {}
    for key, value in old.items():
        if key.startswith('Item') and not value['id']:
            ids.update({new[key]['payload']['item']['labels']['en']: new[key]['id']})
    return ids

def get_discipline(answers):
    ids = []
    md = 0
    nmd = 0
    for key in answers.get('processstep', []):
        for key2 in answers['processstep'][key].get('discipline', []):
            if answers['processstep'][key]['discipline'][key2].get('ID') and answers['processstep'][key]['discipline'][key2]['ID'] not in ids:
                if 'mardi' in answers['processstep'][key]['discipline'][key2]['ID'] or 'wikidata' in answers['processstep'][key]['discipline'][key2]['ID']:
                    answers.setdefault('nonmathdiscipline', {}).update({nmd: {'ID': answers['processstep'][key]['discipline'][key2]['ID'],
                                                                              'Name': answers['processstep'][key]['discipline'][key2]['Name']}})
                    nmd += 1
                    ids.append(answers['processstep'][key]['discipline'][key2]['ID'])
                elif 'msc' in answers['processstep'][key]['discipline'][key2]['ID']:
                    answers.setdefault('mathsubject', {}).update({md: {'ID': answers['processstep'][key]['discipline'][key2]['ID'],
                                                                       'Name': answers['processstep'][key]['discipline'][key2]['Name']}})
                    md += 1
                    ids.append(answers['processstep'][key]['discipline'][key2]['ID'])
    return answers





'''Module containing Providers for the Publication Documentation'''

from rdmo.options.providers import Provider
from ..queries import query_sources

class Publication(Provider):
    '''Publication Provider (MaRDI Portal / Wikidata),
       No User Creation, Refresh Upon Selection
    '''

    search = True
    refresh = True

    def get_options(self, project, search=None, user=None, site=None):
        '''Queries MaRDI Portal, Wikidata, MathAlgoDB for user input'''
        if not search or len(search) < 3:
            return []

        # Define the query parameter
        query_id = 'algorithm/queries/provider_publication.sparql'
        sources = ['mathalgodb', 'mardi', 'wikidata']

        return query_sources(search, query_id, sources)



'''Main Export Functions of MaRDMO'''

import logging
from abc import ABC

import requests

from django import forms
from django.conf import settings
from django.http import HttpRequest
from django.shortcuts import redirect, render, reverse
from django.utils.translation import gettext_lazy as _

from rdmo.projects.exports import Export
from rdmo.services.providers import OauthProviderMixin

from .config import endpoint
from .getters import (
    get_answers,
    get_general_item_url,
    get_mathmoddb,
    get_mathalgodb,
    get_options,
    get_questions,
    get_sparql_query
)
from .helpers import  (
    compare_items,
    inline_mathml,
    merge_dicts_with_unique_keys,
    process_question_dict
)
from .oauth2 import OauthProviderMixin

from .model.worker import PrepareModel
from .model.checks import Checks

from .algorithm.worker import algorithm_relations
from .algorithm.utils import (
    dict_to_triples_mathalgodb,
    generate_sparql_insert_with_new_ids_mathalgodb,
    update_ids
)

from .workflow.utils import get_discipline
from .workflow.worker import prepareWorkflow
from .search.worker import search
from .publication.worker import PublicationRetriever

logger = logging.getLogger(__name__)

class BaseMaRDMOExportProvider(OauthProviderMixin, Export, ABC):
    '''Base Class providing URLs and Credentials.'''
    @property
    def oauth2_client_id(self):
        '''Provide OAuth2 Client ID'''
        return settings.MARDMO_PROVIDER['oauth2_client_id']

    @property
    def oauth2_client_secret(self):
        '''Provide OAuth2 Client Secret'''
        return settings.MARDMO_PROVIDER['oauth2_client_secret']

    @property
    def mathalgodb_id(self):
        '''Provide MathAlgoDB ID'''
        return settings.MARDMO_PROVIDER['mathalgodb_id']

    @property
    def mathalgodb_secret(self):
        '''Provide MathAlgoDB Secret'''
        return settings.MARDMO_PROVIDER['mathalgodb_secret']

    @property
    def wikibase_url(self):
        '''Provide MaRDI Portal URL'''
        return endpoint['mardi']['uri']

    @property
    def authorize_url(self):
        '''Provide OAuth2 Authorize URL'''
        return f'{self.wikibase_url}/w/rest.php/oauth2/authorize'

    @property
    def token_url(self):
        '''Provide OAuth2 Token URL'''
        return f'{self.wikibase_url}/w/rest.php/oauth2/access_token'

    @property
    def redirect_path(self):
        '''Provide OAuth2 Callback Path'''
        return reverse('oauth_callback', args=['wikibase'])

    def get_authorize_params(self, request, state):
        '''Get Authoritaion Parameter for OAuth2 Protocol'''
        return {
            'response_type': 'code',
            'client_id': self.oauth2_client_id,
            'state': state
        }

    def get_callback_data(self, request):
        '''Get Callback Data for OAuth2 Protocol'''
        return {
            'client_id': self.oauth2_client_id,
            'client_secret': self.oauth2_client_secret,
            'grant_type': 'authorization_code',
            'code': request.GET.get('code')
        }

class MaRDMOExportProvider(BaseMaRDMOExportProvider):
    """Handles MaRDMO export logic.

    Attributes:
        request: The Django HttpRequest associated with the provider.
    """

    request: HttpRequest

    class ExportForm(forms.Form):
        '''Placeholder Export Form Class'''

    def render(self):
        '''Render Preview of User Answers'''
        # MaRDMO: Mathematical Model Documentation
        if str(self.project.catalog).endswith('mardmo-model-catalog'):

            # Get answers, options, and mathmoddb
            answers, options, mathmoddb = self.get_post_data()

            # Adjust MathML for Preview
            inline_mathml(answers)

            return render(
                self.request,
                'MaRDMO/mardmoPreview.html', 
                {
                    'form': self.ExportForm(),
                    'include_file': 'MaRDMO/modelTemplate.html',
                    'include_params': {
                        'title': self.project.title
                    },
                    'answers': answers,
                    'option': options|mathmoddb
                },
                status=200
            )

        # MaRDMO: Algorithm Documentation
        if str(self.project.catalog).endswith('mardmo-algorithm-catalog'):

            # Get answers, options, mathalgodb
            answers, options, mathalgodb = self.get_post_data()

            return render(
                self.request,
                'MaRDMO/mardmoPreview.html',
                {
                    'form': self.ExportForm(),
                    'include_file': 'MaRDMO/algorithmTemplate.html',
                    'include_params': {
                        'title': self.project.title
                    },
                    'answers': answers,
                    'option': options|mathalgodb
                },
                status=200
            )

        # MaRDMO: Search Interdisciplinary Workflows, Mathematical Models or Algorithms
        if str(self.project.catalog).endswith('mardmo-search-catalog'):

            answers, options = self.get_post_data()

            return render(
                self.request,
                'MaRDMO/mardmoPreview.html',
                {
                    'form': self.ExportForm(),
                    'include_file': 'MaRDMO/searchTemplate.html',
                    'include_params': {
                        'title': self.project.title
                    },
                    'answers': answers,
                    'option': options
                },
                status=200
            )

        # MaRDMO: Interdisciplinary Workflow Documentation
        if str(self.project.catalog).endswith('mardmo-interdisciplinary-workflow-catalog'):

            answers, options = self.get_post_data()

            # Get Model Data
            prepare = prepareWorkflow()
            answers = prepare.preview(answers)

            # Adjust MathML for Preview
            inline_mathml(answers)

            return render(
                self.request,
                'MaRDMO/mardmoPreview.html',
                {
                    'form': self.ExportForm(),
                    'include_file': 'MaRDMO/workflowTemplate.html',
                    'include_params': {
                        'title': self.project.title
                    },
                    'answers': answers,
                    'option': options,
                    'mathmoddbURI': endpoint['mardi']['uri'],
                    'mathalgodbURI': endpoint['mathalgodb']['uri'],
                    'mardiURI': get_general_item_url(),
                    'wikidataURI': endpoint['wikidata']['uri']
                },
                status=200
            )

        # Non-MaRDMO Catalog
        return render(
            self.request,
            'core/error.html',
            {
                'title': _('Catalog Error'),
                'errors': [_('The catalog is not supported by the MaRDMO Plugin.')]
            },
            status=200
        )

    def submit(self):
        """Dispatch submission based on catalog type."""

        # Handle cancel for all submissions
        if 'cancel' in self.request.POST:
            return redirect('project', self.project.id)

        catalog = str(self.project.catalog)

        if catalog.endswith('mardmo-model-catalog'):
            return self.submit_mardmo_model()
        if catalog.endswith('mardmo-algorithm-catalog'):
            return self.submit_mardmo_algorithm()
        if catalog.endswith('mardmo-search-catalog'):
            return self.submit_mardmo_search()
        if catalog.endswith('mardmo-interdisciplinary-workflow-catalog'):
            return self.submit_mardmo_workflow()

        # Default fallback if catalog type is unknown
        return render(
            self.request,
            'core/error.html',
            {
                'title': _('Unknown catalog'),
                'errors': [_('Cannot handle this catalog type.')]
            },
            status=400
        )

    def submit_mardmo_model(self):
        """Submit MaRDMO Mathematical Model Documentation."""

        # Check MaRDI Portal Credentials
        if not (self.oauth2_client_id and self.oauth2_client_secret):
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _('Missing Credentials'),
                    'errors': [_('Credentials for MaRDI Portal are missing!')]
                },
                status=200
            )

        answers, *__ = self.get_post_data()

        # Validate documentation completeness / consistency
        checker = Checks()
        err = checker.run(self.project, answers)
        if err:
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _("Incomplete or Inconsistent Documentation"),
                    'errors': err
                },
                status=200
            )

        # Prepare payload
        try:
            prepare = PrepareModel()
            payload = prepare.export(answers, self.wikibase_url)
        except (ValueError, KeyError) as err:
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _('Value Error'),
                    'errors': [err]
                },
                status=200
            )

        return self.post(self.request, payload)


    def submit_mardmo_algorithm(self):
        """Submit MaRDMO Algorithm Documentation."""

        # Check MathAlgoDB Credentials
        if not (self.mathalgodb_id and self.mathalgodb_secret):
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _('Missing Credentials'),
                    'errors': [_('Credentials for MathAlgoDB are missing!')]
                },
                status=200
            )

        answers, *__ = self.get_post_data()

        # Merge answers related to mathematical model
        merged_dict = merge_dicts_with_unique_keys(
            answers,
            [
                'algorithm',
                'problem',
                'software',
                'benchmark',
                'publication'
            ]
        )

        # Generate list of triples
        triple_list, ids = dict_to_triples_mathalgodb(merged_dict)

        # Generate SPARQL query
        query = generate_sparql_insert_with_new_ids_mathalgodb(triple_list)

        try:
            response = requests.post(
                endpoint['mathalgodb']['update'],
                data=query,
                headers={
                    "Content-Type": "application/sparql-update",
                    "Accept": "text/turtle"
                },
                auth=(self.mathalgodb_id, self.mathalgodb_secret),
                verify=False,
                timeout=60
            )
        except requests.RequestException as err:
            return render(
                self.request,
                'core/error.html',
                {'title': _('Request Error'), 'errors': [err]},
                status=500
            )

        if response.status_code == 204:
            ids = update_ids(
                self.project,
                ids,
                get_sparql_query('algorithm/queries/id_check.sparql'),
                endpoint['mathalgodb']['sparql'],
                'mathalgodb'
            )
            return render(
                self.request,
                'MaRDMO/algorithmExport.html',
                {'ids': ids, 'mathalgodb_uri': endpoint['mathalgodb']['uri']},
                status=200
            )

        return render(
            self.request,
            'MaRDMO/workflowError.html',
            {'error': 'The algorithm could not be integrated into the MathAlgoDB!'},
            status=200
        )

    def submit_mardmo_search(self):
        """Submit MaRDMO Search Interdisciplinary Workflow, Models, or Algorithms."""

        answers, options = self.get_post_data()

        if answers['search']['options'] == options['InterdisciplinaryWorkflow']:
            datatype = "Workflow(s)"
            source = "MaRDI Portal"
        elif answers['search']['options'] == options['MathematicalModel']:
            datatype = "Mathematical Model(s)"
            source = "MathModDB KG"
        elif answers['search']['options'] == options['Algorithm']:
            datatype = "Algorithm(s)"
            source = "MathAlgoDB KG"
        else:
            datatype = "Unknown"
            source = "Unknown"

        return render(
            self.request,
            'MaRDMO/searchResults.html',
            {
                'datatype': datatype,
                'source': source,
                'noResults': answers['no_results'],
                'links': answers['links']
            },
            status=200
        )

    def submit_mardmo_workflow(self):
        """Submit MaRDMO Interdisciplinary Workflow Documentation."""

        # Check MaRDI Portal Credentials
        if not (self.oauth2_client_id and self.oauth2_client_secret):
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _('Missing Credentials'),
                    'errors': [_('Credentials for MaRDI Portal are missing!')]
                },
                status=200
            )

        answers, *__ = self.get_post_data()

        try:
            prepare = prepareWorkflow()
            payload = prepare.export(answers, self.project.title, self.wikibase_url)
        except (ValueError, KeyError) as err:
            return render(
                self.request,
                'core/error.html',
                {
                    'title': _('Value Error'),
                    'errors': [err]
                },
                status=200
            )

        return self.post(self.request, payload)

    def post_success(self, request, init, final):
        '''Display created Items, if Post successful'''
        ids = compare_items(init, final)

        # Links to newly created Entities
        return render(
            request,
            'MaRDMO/portalExport.html',
            {
                'ids': ids,
                'mardi_uri': get_general_item_url
            },
            status=200
        )

    def get_post_data(self):
        '''Function which gathers User Answers for individual catalogs.'''
        options = get_options()

        # MaRDMO: Mathematical Model Documentation
        if str(self.project.catalog).endswith('mardmo-model-catalog'):

            # Load Data for Model & Publication Documentation
            questions = get_questions('model') | get_questions('publication')
            mathmoddb = get_mathmoddb()

            answers = process_question_dict(
                project = self.project,
                questions = questions,
                get_answer = get_answers
            )

            # Retrieve Publications related to Model
            publication = PublicationRetriever()
            answers = publication.workflow_or_model(
                project = self.project,
                snapshot = self.snapshot,
                answers = answers,
                options = options
            )

            # Prepare Mathematical Model (Preview)
            prepare = PrepareModel()
            answers = prepare.preview(answers)

            return answers, options, mathmoddb

        # MaRDMO: Algorithm Documentation
        if str(self.project.catalog).endswith('mardmo-algorithm-catalog'):

            # Load Data for Mathematical Model Documentation
            questions = get_questions('algorithm') | get_questions('publication')
            mathalgodb = get_mathalgodb()

            answers = process_question_dict(
                project = self.project,
                questions = questions,
                get_answer = get_answers
            )

            # Refine Mathematical Model Information
            answers = algorithm_relations(answers)

            # Retrieve Publications related to Workflow
            publication = PublicationRetriever()
            answers = publication.algorithm(
                project = self.project,
                answers = answers
            )

            return answers, options, mathalgodb

        # MaRDMO: Search Interdisciplinary Workflow, Mathematical Model or Algorithm
        if str(self.project.catalog).endswith('mardmo-search-catalog'):

            # Load Data for Interdisciplinary Workflow, Mathematical Model or Algorithm Search
            questions = get_questions('search')

            answers = process_question_dict(
                project = self.project,
                questions = questions,
                get_answer = get_answers
            )

            # Get Results from MaRDI Resources
            answers = search(
                answers,
                options
            )

            return answers, options

        # MaRDMO: Interdisciplinary Workflow Documentation
        if str(self.project.catalog).endswith('mardmo-interdisciplinary-workflow-catalog'):

            # Load Data for Interdisciplinary Workflow Documentation
            questions = get_questions('workflow') | get_questions('publication')

            answers = process_question_dict(
                project = self.project,
                questions = questions,
                get_answer = get_answers
            )

            # Refine associated Disciplines
            answers = get_discipline(answers)

            # Retrieve Publications related to Workflow
            publication = PublicationRetriever()
            answers = publication.workflow_or_model(
                project = self.project,
                snapshot = self.snapshot,
                answers = answers,
                options = options
            )

            return answers, options

        # Default fallback if catalog type is unknown
        return render(
            self.request,
            'core/error.html',
            {
                'title': _('Unknown catalog'),
                'errors': [_('Cannot handle this catalog type.')]
            },
            status=400
        )



'''Module containing Handlers for the Model Documentation'''

from rdmo.options.models import Option

from . import models

from .constants import props, relatant_uris, relation_uris, index_counters

from ..config import BASE_URI, endpoint
from ..getters import get_items, get_mathmoddb, get_properties, get_questions, get_sparql_query
from ..helpers import value_editor
from ..queries import query_sparql
from ..adders import (
    add_basics,
    add_entities,
    add_properties,
    add_relations_static,
    add_relations_flexible,
    add_references
)

class Information:
    '''Class containing functions, querying external sources for specific
       entities and integrating the related metadata into the questionnaire.'''

    def __init__(self):
        # Load shared data once
        self.questions = get_questions("model") | get_questions('publication')
        self.mathmoddb = get_mathmoddb()
        self.base = BASE_URI

    def field(self, instance):
        '''Research Field Information'''

        # Research Field specific Questions.
        field = self.questions["Research Field"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Informatiom
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Research Field',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/field.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results
        data = models.ResearchField.from_query(results)

        # Add long Descriptions (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{field["Long Description"]["uri"]}',
                info = {
                    'text': description_long, 
                    'collection_index': idx,
                    'set_index': 0, 
                    'set_prefix': instance.set_index
                }
            )

        # Add Research Field Relations (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['Field'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{field["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{field["IntraClassElement"]["uri"]}',
            },
        )

        # Add Publications
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def problem(self, instance):
        '''Research Problem Information'''

        # Research Problem specific Questions.
        problem = self.questions["Research Problem"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Research Problem',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/problem.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results
        data = models.ResearchProblem.from_query(results)

        # Add long Description (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{problem["Long Description"]["uri"]}',
                info = {
                    'text': description_long,
                    'collection_index': idx, 
                    'set_index': 0, 
                    'set_prefix': instance.set_index
                }
            )

        # Add Research Field Relations (static)
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': props['RP2RF']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{self.base}{problem["RFRelatant"]["uri"]}'
            }
        )

        # Add Research Problem Relations (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['Problem'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{problem["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{problem["IntraClassElement"]["uri"]}',
            },
        )

        # Add Publications
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def quantity(self, instance):
        '''Quantity [Kind] Information'''

        # Quantity specific Questions.
        quantity = self.questions["Quantity"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Quantity',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/quantity.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results
        data = models.QuantityOrQuantityKind.from_query(results)

        # Add long Description (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{quantity["Long Description"]["uri"]}',
                info = {
                    'text': description_long,
                    'collection_index': idx, 
                    'set_index': 0, 
                    'set_prefix': instance.set_index
                }
            )

        # Add Type
        if data.qclass:
            value_editor(
                project = instance.project,
                uri = f'{self.base}{quantity["QorQK"]["uri"]}',
                info = {
                    'option': Option.objects.get(
                        uri = self.mathmoddb[results[0]['class']['value']]
                    ),
                    'set_index': instance.set_index
                }
            )

        # Add Properties
        add_properties(
            project = instance.project,
            data = data,
            uri = f'{self.base}{quantity["Properties"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add References
        add_references(
            project = instance.project,
            data = data,
            uri = f'{self.base}{quantity["Reference"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add defining Formula(s)
        for idx, formula in enumerate(data.formulas):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{quantity["Formula"]["uri"]}',
                info = {
                    'text': formula, 
                    'collection_index': idx, 
                    'set_index': 0, 
                    'set_prefix': f"{instance.set_index}|0"
                }
            )

        # Add Symbol(s)
        for idx, symbol in enumerate(data.symbols):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{quantity["Element Symbol"]["uri"]}',
                info = {
                    'text': symbol,
                    'set_index': idx, 
                    'set_prefix': f"{instance.set_index}|0|0"
                }
            )

        # Add Quantities
        for idx, encoded_quantity in enumerate(data.contains_quantity):
            source, _ = encoded_quantity.id.split(':')
            value_editor(
                project = instance.project,
                uri = f'{self.base}{quantity["Element Quantity"]["uri"]}',
                info = {
                    'text': f"{encoded_quantity.label} ({encoded_quantity.description}) [{source}]",
                    'external_id': encoded_quantity.id,
                    'set_index': idx,
                    'set_prefix': f"{instance.set_index}|0|0"
                }
            )

        # Add Quantity [Kind] Relations (flexible)
        for prop in props['Quantity']:
            for value in getattr(data, prop):
                qclass_pair = (data.qclass, value.item_class)
                value_editor(
                    project=instance.project,
                    uri=f"{self.base}{quantity[relation_uris[qclass_pair]]['uri']}",
                    info = {
                        'option': Option.objects.get(uri = self.mathmoddb[prop]),
                        'set_index': index_counters[qclass_pair],
                        'set_prefix': f"{instance.set_index}"
                    }
                )
                value_editor(
                    project=instance.project,
                    uri=f"{self.base}{quantity[relatant_uris[qclass_pair]]['uri']}",
                    info = {
                        'text': f"{value.label} ({value.description}) [{source}]",
                        'external_id': value.id,
                        'set_index': index_counters[qclass_pair],
                        'set_prefix': f"{instance.set_index}"
                    }
                )
                index_counters[qclass_pair] += 1

        # Add Publications
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def formulation(self, instance):
        '''Mathematical Formulation Information'''

        # Formulation specific Questions.
        formulation = self.questions["Mathematical Formulation"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Mathematical Formulation',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/formulation.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results
        data = models.MathematicalFormulation.from_query(results)

        # Add long Description (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{formulation["Long Description"]["uri"]}',
                info = {
                    'text': description_long,
                    'collection_index': idx,
                    'set_index': 0,
                    'set_prefix': instance.set_index
                }
            )

        # Add Properties
        add_properties(
            project = instance.project,
            data = data,
            uri = f'{self.base}{formulation["Properties"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Formula(s)
        for idx, formula in enumerate(data.formulas):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{formulation["Formula"]["uri"]}',
                info = {
                    'text': formula, 
                    'collection_index': idx, 
                    'set_index': 0, 
                    'set_prefix': f"{instance.set_index}|0"
                }
            )

        # Add Symbol(s)
        for idx, symbol in enumerate(data.symbols):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{formulation["Element Symbol"]["uri"]}',
                info = {
                    'text': symbol, 
                    'set_index': idx, 
                    'set_prefix': f"{instance.set_index}|0|0"
                }
            )

        # Add Quantities
        for idx, quantity in enumerate(data.contains_quantity):
            source, _ = quantity.id.split(':')
            value_editor(
                project = instance.project,
                uri = f'{self.base}{formulation["Element Quantity"]["uri"]}',
                info = {
                    'text': f"{quantity.label} ({quantity.description}) [{source}]", 
                    'external_id': quantity.id, 
                    'set_index': idx, 
                    'set_prefix': f"{instance.set_index}|0|0"
                }
            )

        # Add Formulation Relations I (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['MF2MF'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': f"{instance.set_index}|0"
            },
            statement = {
                'relation': f'{self.base}{formulation["MF2MF"]["uri"]}',
                'relatant': f'{self.base}{formulation["MFRelatant"]["uri"]}',
            },
        )

        # Add Formulation Relations II (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['Formulation'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{formulation["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{formulation["IntraClassElement"]["uri"]}',
                'assumption': f'{self.base}{formulation["Assumption"]["uri"]}',
            },
        )

        # Add Publications
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def task(self, instance):
        '''Task Information'''

        # Task specific Questions.
        task = self.questions["Task"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Task',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/task.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        #Structure Results
        data = models.Task.from_query(results)

        # Add long Description (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{self.base}{task["Long Description"]["uri"]}',
                info = {
                    'text': description_long,
                    'collection_index': idx,
                    'set_index': 0,
                    'set_prefix': instance.set_index
                }
            )

        # Add Properties
        add_properties(
            project = instance.project,
            data = data,
            uri = f'{self.base}{task["Properties"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Formulations Relations (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['T2MF'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': f"{instance.set_index}|0"
            },
            statement = {
                'relation': f'{self.base}{task["T2MF"]["uri"]}',
                'relatant': f'{self.base}{task["MFRelatant"]["uri"]}',
            },
        )

        # Add Quantity Relations (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['T2Q'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': f"{instance.set_index}|0"
            },
            statement = {
                'relation': f'{self.base}{task["T2Q"]["uri"]}',
                'relatant': f'{self.base}{task["QRelatant"]["uri"]}',
            },
        )

        # Add Task Relations (flexible)
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['Task'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{self.base}{task["IntraClassRelation"]["uri"]}',
                'relatant': f'{self.base}{task["IntraClassElement"]["uri"]}',
                'assumption': f'{self.base}{task["Assumption"]["uri"]}',
                'order': f'{self.base}{task["Order Number"]["uri"]}',
            },
        )

        # Add Publications
        add_entities(
            project = instance.project,
            question_set = f'{self.base}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )

    def model(self, instance):
        '''Mathematical Model Information'''

        # Model specific Questions.
        model = self.questions["Mathematical Model"]

        # Stop if no Text or 'not found' in ID Field
        if not instance.text or instance.text == 'not found':
            return

        # Add basic Information
        add_basics(
            project = instance.project,
            text = instance.text,
            questions = self.questions,
            item_type = 'Mathematical Model',
            index = (0, instance.set_index)
        )

        # Get source and ID of Item
        source, identifier = instance.external_id.split(':')

        # Only consider MaRDI (so far)
        if source != 'mardi':
            return

        # If Item from MathModDB, set up Query and...
        query = get_sparql_query('model/queries/model.sparql').format(
            identifier,
            **get_items(),
            **get_properties()
        )

        # ...get Results.
        results = query_sparql(query, endpoint[source]['sparql'])

        # Stop if no Results retrieved from external source
        if not results:
            return

        # Structure Results
        data = models.MathematicalModel.from_query(results)

        # Add long Description (optional)
        for idx, description_long in enumerate(data.description_long):
            value_editor(
                project = instance.project,
                uri = f'{BASE_URI}{model["Long Description"]["uri"]}',
                info = {
                    'text': description_long,
                    'collection_index': idx, 
                    'set_index': 0,
                    'set_prefix': instance.set_index
                }
            )

        # Add Properties
        add_properties(
            project = instance.project,
            data = data,
            uri = f'{BASE_URI}{model["Properties"]["uri"]}',
            set_prefix = instance.set_index
        )

        # Add Research Problems Relations (static)
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': props['MM2RP']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{BASE_URI}{model["RPRelatant"]["uri"]}'
            }
        )

        # Add Formulations contained in Mathematical Model
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['MM2MF'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': f"{instance.set_index}|0"
            },
            statement = {
                'relation': f'{BASE_URI}{model["MM2MF"]["uri"]}',
                'relatant': f'{BASE_URI}{model["MFRelatant"]["uri"]}',
                'order': f'{BASE_URI}{model["Order Number"]["uri"]}',
            },
        )

        # Add Relations between Mathematical Models and Tasks to Questionnaire
        add_relations_static(
            project = instance.project,
            data = data,
            props = {
                'keys': props['MM2T']
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relatant': f'{BASE_URI}{model["TRelatant"]["uri"]}'
            }
        )

        # Add related Models to questionnaire
        add_relations_flexible(
            project = instance.project,
            data = data,
            props = {
                'keys': props['Model'],
                'mapping': self.mathmoddb,
            },
            index = {
                'set_prefix': instance.set_index
            },
            statement = {
                'relation': f'{BASE_URI}{model["IntraClassRelation"]["uri"]}',
                'relatant': f'{BASE_URI}{model["IntraClassElement"]["uri"]}',
                'assumption': f'{BASE_URI}{model["Assumption"]["uri"]}',
            },
        )

        # Add Publications to Questionnaire
        add_entities(
            project = instance.project,
            question_set = f'{BASE_URI}{self.questions["Publication"]["uri"]}',
            datas = data.publications,
            source = source,
            prefix = 'P'
        )
